/* libhtmlparse: Quick and simple HTML parser
   prettyHTML.c: Demonstration application
   Mooneer Salem (mooneer@translator.cx) */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "libhtmlparse.h"

int numTabs;
FILE *output;

void printTabs () {
	int i;
	for (i=1; i <= numTabs; i++) fprintf(output, "  ");
}

int comments (char *comment) {
	fprintf(output, "%s", comment);
	return 0;
}

int texts (char *text) {
	fprintf(output, "%s", text);
	return 0;
}

int endtag (char *tag) {
	numTabs--;
	printTabs();
	fprintf(output, "</%s>\n", tag);

	return 0;
}

int starttag (char *tag, /*@null@*/ struct ArgvTable *args, int numArgs) {
	int i;

	printTabs();
	fprintf(output, "<%s", tag);
	if (args) {
		for(i=0; i<numArgs; i++) {
			if (args[i].val) {
				fprintf(output, " %s=\"%s\"", args[i].arg, args[i].val);
			} else {
				fprintf(output, " %s", args[i].arg);
			}
		}
	}
	fprintf(output, ">\n");

	numTabs++;

	return 0;
}

int decltag (char *tag, /*@null@*/ struct ArgvTable *args, int numargs) {
        int i;

        printTabs();
        fprintf(output, "<!%s", tag);
        if (args) {
                for(i=0; i<numargs; i++) {
                        if (args[i].val) {
                                fprintf(output, " %s=%s ", args[i].arg, args[i].val);
                        } else {
				if (strchr(args[i].arg, ' ')) {
					fprintf(output, " \"%s\"", args[i].arg);
				} else {
                                	fprintf(output, " %s", args[i].arg);
				}
                        }
                }
        }
        fprintf(output, ">\n");
        
        return 0;
}

int commentStart () {
	printTabs();
	fprintf(output, "<!--");
	return 0;
}

int commentEnd () {
	fprintf(output, "-->\n");
	return 0;
}

int textStart () {
	printTabs();
	return 0;
}

int textEnd () {
	fprintf(output, "\n");
	return 0;
}

void open_error () {
	perror("open");
	exit(EXIT_FAILURE);
}

void usage (char **argv) {
	fprintf(stderr, "Usage: %s [source] [destination]\n", argv[0]);
	exit(0);
}

int phpcallback (char *text) {
	fprintf(output, "<?php %s ?>", text);
	return 0;
}

int main (int argc, char **argv) {
	char *html;
	FILE *fp;
	struct stat filer;

	if (argc < 3) {
		usage(argv);
	}

	fp = fopen(argv[1], "r");
	if (!fp) {
		open_error();
		exit(EXIT_FAILURE);
	}

	if (stat(argv[1], &filer) == -1) {
		fprintf(stderr, "error: cannot open %s\n", argv[1]);
		exit(EXIT_FAILURE);
	}

	html = (char*)calloc(filer.st_size + 1, 1);
	if (!html) {
		printf("memory error\n");
		exit(EXIT_FAILURE);
	}

	(void)fread(html, 1, (size_t)filer.st_size, fp);
	(void)fclose(fp);

	output = fopen(argv[2], "w");
	if (!output) {
		free(html);
		html = NULL;
		open_error();
		exit(EXIT_FAILURE);
	}

	commentCallBack = comments;
	textCallBack = texts;
	startCallBack = starttag;
	endCallBack = endtag;
	commentStartCallBack = commentStart;
	commentEndCallBack = commentEnd;
	textStartCallBack = textStart;
	textEndCallBack = textEnd;
	declCallBack = decltag;
	phpCallBack = phpcallback;

	(void)parse(html);
	free(html);

	(void)fclose(output);

	return 0;
}


