/* libhtmlparse: Quick and simple HTML parser
   Mooneer Salem (mooneer@translator.cx) */

#ifndef LIBHTML_H
#define LIBHTML_H

/****************
* from the guts *
*****************/

/* struct holding the arguments of tags */

struct ArgvTable {
	char *arg, *val;
};

/* struct for streamparsing and errorhandling */

struct ErrorTree {
	unsigned int line;
	unsigned char type;
	char *problem;
};

/* obligatory prototypes */

static const char *parseText     	(const char *);
static const char *parseComment  	(const char *);
static const char *parseEndTag   	(const char *);
static const char *parsePHP      	(const char *);
static const char *parseXMLtag   	(const char *);
static const char *eatUp         	(const char *);
static const char *eatUpText     	(const char *);
static const char *parseXML      	(const char *);
static const char *parseStartTag 	(const char *);
static const char *parseDecl     	(const char *);
static const char *parseForEntities	(const char *);

/**********************************************************************/

/***********************************************************************
*                        The CallBacks                                 *
* You may call one ore several or even all callbacks. Except of the    *
* XHTMLCallBack, all CallBacks will work as expected and described     *
*                                                                      *
* XHTMLCallBack:                                                       *
* The XHTMLCallBack is a special case, because you can decide, if the  *
* XHTML specific tags should be handeled as a start- AND endtag, or    *
* as an XHTML tag. If you call nothing, except start and endtag, the   *
* behaviour is, that you'll get a start AND an endtag called back.     *
* If you call XHTMLCallBack, it will only give you the XHTML call back.*
*                                                                      *
* If you are in doubt or simply confused now, call XHTMLCallBack()     *
************************************************************************/

/* handle comments and javascript */
int (*commentCallBack) (char *comment);
int (*commentStartCallBack) ();
int (*commentEndCallBack) ();

/* Declaration e.g. <!DOCTYPE HTML ... */
int (*declCallBack) (char *tag, /*@null@*/ struct ArgvTable *args, int numargs);

/* Start tag e.g. <html>, with arguments, args may be NULL, numargs may be 0 */
int (*startCallBack) (char *tag, /*@null@*/ struct ArgvTable *args, int numargs);

/* End tag e.g. </html>*/
int (*endCallBack) (char *tag);

/* handle plain text */
int (*textCallBack) (char *text);
int (*textStartCallBack) ();
int (*textEndCallBack) ();

/* PHP inserts. BUG(?): if someone prints another PHP function from this PHP function
   our lib will get confused. */
int (*phpCallBack) (char *text);

/* empty tags like <hr/>, <br/>, with arguments, args may be NULL, numargs may be 0 */
int (*XHTMLCallBack) (char *tag, /*@null@*/ struct ArgvTable *args, int numargs);

/* XML tags <?xml>, with arguments, args may be NULL, numargs may be 0 */
int (*xmlCallBack) (char *tag, /*@null@*/ struct ArgvTable *args, int numargs);

/* entities like &auml;,&#228; text will inherit all chars between '&' and ';' */

int (*entityCallBack) (char *text);

/* the script inside <script> ... </script> */
/* int (*scriptCallBack) (char *text);  */

/* parse the inputfile */
/*@null@*/ struct ErrorTree *parse (const char *html);

/**********************************************************************/

/* more of the obligatory prototypes ( one should clean up here! ;-) */

/* puts found arguments into the argument struct */
static struct ArgvTable *addArgToTable (/*@null@*/ struct ArgvTable *args, char *arg, /*@null@*/ char *val);

/* cleans the argument struct */
static void freeArgs (/*@null@*/ struct ArgvTable *args);

#endif /* LIBHTML_H */

